from osgeo import gdal, osr
from math import ceil, floor
import logging

gdal.SetConfigOption("GDAL_CACHEMAX", "1000")
logger = logging.getLogger("preprocessor")

GDT_SIZES = {
    # geotiff data types and their respective sizes in bits
    "Byte": 8,
    "Int16": 16,
    "UInt16": 16,
    "CInt16": 16,
    "Int32": 32,
    "UInt32": 32,
    "CInt32": 32,
    "Float32": 32,
    "CFloat32": 32,
    "Float64": 64,
    "CFloat64": 64
}


def open_gdal_dataset(input_file):
    logger.debug("Opening file using GDAL: %s" % input_file)
    return gdal.Open(input_file)


def apply_rpc(src_ds):
    # if RPC metadata on source is present, apply it as geotransform using warp through temporary vrt
    rpc = src_ds.GetMetadata("RPC")
    if rpc:
        logger.debug("Applying RPC metadata.")
        dst_ds = gdal.Warp("", src_ds, dstSRS="EPSG:4326", format="VRT", rpc=True, multithread=True, warpMemoryLimit=1024*1024*1024, resampleAlg=gdal.GRA_NearestNeighbour)
        return dst_ds
    return src_ds


def set_rpc_metadata(src_ds, dst_ds):
    # sets RPC metadata from one source dataset to destination dataset if present
    rpc = src_ds.GetMetadata('RPC')
    if rpc:
        dst_ds.SetMetadata(rpc, 'RPC')
    return dst_ds


def write_gdal_dataset(src_ds, driver_name, output_file_name, creation_options=[]):
    # writes dataset to an output file using a given driver and array of creation options (-CO)
    # returns filelist as an array (for later reference)
    driver_instance = gdal.GetDriverByName(driver_name)
    dst_ds = driver_instance.CreateCopy(output_file_name, src_ds, strict=0, options=creation_options)
    dst_ds = None   # write to disk
    return [output_file_name]


def write_gdal_dataset_split(src_ds, driver_name, output_file_name, creation_options=[], split_parts=1):
    # writes dataset to an output file using a given driver and array of creation options (-CO)
    # returns a list of created files
    if driver_name == "COG":
        # COG driver does not have Create method, need to create in memory raster first and then CreateCopy
        driver_instance = gdal.GetDriverByName("MEM")
        creation_options_valid = []  # creation options applied only to result driver
    else:
        driver_instance = gdal.GetDriverByName(driver_name)
        creation_options_valid = creation_options
    logger.info("Writing file to disk. %s" % ("Splitting into %s stripes." % split_parts if split_parts > 1 else ""))
    if split_parts > 1:
        # get image properties
        dst_filenames = []
        y_coord = 0
        cols = src_ds.RasterXSize
        rows = src_ds.RasterYSize
        bands = src_ds.RasterCount
        data_type = src_ds.GetRasterBand(1).DataType
        geo_transform = src_ds.GetGeoTransform()
        projection = osr.SpatialReference()
        projection.ImportFromWkt(src_ds.GetProjectionRef())
        if not geo_transform:
            # provide some default when not set, which will not make sense
            # but at least split images will not be stacked on top of each other
            geo_transform = [0.0, 1.0, 0.0, 0.0, 0.0, -1.0]
        y_origin = geo_transform[3]  # uly
        pixelHeight = geo_transform[5]  # yres
        size = floor(rows / split_parts)  # height of one strip
        for i in range(split_parts):
            # add underscore and padded index to between filename and extension if there is one
            output_file_name_split = output_file_name.rsplit(".", 1)
            output_file_name_crop = "%s_%04d.%s" % (output_file_name_split[0], i, output_file_name_split[1] if len(output_file_name_split) > 1 else "")
            logger.debug("Creating in memory dataset stripe %04d." % i)
            if i < split_parts - 1:
                # all stripes except the last one
                # create a new dataset
                dst_ds = driver_instance.Create(output_file_name_crop, xsize=cols, ysize=size, bands=bands, eType=data_type, options=creation_options_valid)
                # fill bands with data
                for band in range(1, bands + 1):
                    data = src_ds.GetRasterBand(band).ReadAsArray(0, y_coord, cols, size)  # xoff, yoff, xsize, ysize
                    dst_ds.GetRasterBand(band).WriteArray(data)
                    no_data_value = src_ds.GetRasterBand(band).GetNoDataValue()
                    if no_data_value:
                        dst_ds.GetRasterBand(band).SetNoDataValue(no_data_value)
                # move y_start for next iteration
                y_coord += size
            else:
                # for last stripe ysize will be larger by remainder after division by split_parts
                dst_ds = driver_instance.Create(output_file_name_crop, xsize=cols, ysize=size + (rows % split_parts), bands=bands, eType=data_type, options=creation_options_valid)
                for band in range(1, bands + 1):
                    data = src_ds.GetRasterBand(band).ReadAsArray(0, y_coord, cols, size + rows % split_parts)  # xoff, yoff, xsize, ysize
                    dst_ds.GetRasterBand(band).WriteArray(data)
                    no_data_value = src_ds.GetRasterBand(band).GetNoDataValue()
                    if no_data_value:
                        dst_ds.GetRasterBand(band).SetNoDataValue(no_data_value)
            # set new geotransform
            new_y = y_origin + (i * size * pixelHeight)
            dst_ds.SetGeoTransform((geo_transform[0], geo_transform[1], geo_transform[2], new_y, geo_transform[4], geo_transform[5]))
            dst_ds.SetProjection(projection.ExportToWkt())
            logger.debug("Writing stripe file %04d." % i)
            if driver_name == "COG":
                cog_driver = gdal.GetDriverByName("COG")
                if cog_driver:
                    output_ds = cog_driver.CreateCopy(output_file_name_crop, dst_ds, options=creation_options)
                    output_ds = None  # write to disk
                else:
                    raise Exception("COG driver was not found. Please upgrade GDAL to version >= 3.1")
            else:
                dst_ds = None  # write to disk
            dst_filenames.append(output_file_name_crop)
    else:
        # do not split, just create output
        dst_filenames = write_gdal_dataset(src_ds, driver_name, output_file_name, creation_options)
    logger.info("Finished writing %s files to disk." % split_parts)
    return dst_filenames


def correct_geo_transform(src_dst):
    # input - gdal dataset
    # sets new geotransform if necessary by creating control points of a raster with switched height and width - by Petr
    # returns - gdal dataset
    ulx, xres, xskew, uly, yskew, yres = src_dst.GetGeoTransform()
    # test geotransform if necessary to shift
    if xres == 0.0 and yres == 0.0:
        logger.debug("Malformed geotransform xres,yres=0 detected, correcting.")
        # malformed image, compute xres and yres switched in geotransform
        lrx = ulx + (src_dst.RasterXSize * xskew)
        lry = uly + (src_dst.RasterYSize * yskew)
        # [ulx, lrx, lry, uly] - bounds = lon_min, lon_max, lat_min, lat_max
        fp = [[0, src_dst.RasterXSize, src_dst.RasterXSize, 0], [0, 0, src_dst.RasterYSize, src_dst.RasterYSize]]
        tp = [[ulx, lrx, lrx, ulx], [lry, lry, uly, uly]]
        pix = list(zip(fp[0], fp[1]))
        coor = list(zip(tp[0], tp[1]))
        # compute the gdal.GCP parameters
        gcps = []
        for index, txt in enumerate(pix):
            gcps.append(gdal.GCP())
            gcps[index].GCPPixel = pix[index][0]
            gcps[index].GCPLine = src_dst.RasterYSize - int(pix[index][1])
            gcps[index].GCPX = coor[index][0]
            gcps[index].GCPY = coor[index][1]
        # get correct geotransform from gcps
        geotransform_new = gdal.GCPsToGeoTransform(gcps)
        # overwrite geotransform with new
        src_dst.SetGeoTransform(geotransform_new)
    return src_dst


def split_check(dst, limit=1024 ** 3):
    # returns number of parts to which split the resulting image to get each part within limit
    # assuming non-compressed resulting image
    parts = 1
    size = 0
    if dst is not None:
        for i in range(1, dst.RasterCount + 1):
            size += dst.RasterXSize * dst.RasterYSize * GDT_SIZES[gdal.GetDataTypeName(dst.GetRasterBand(i).DataType)] / 8
        parts = ceil(size / limit)
    return parts


def create_vrt_dataset(src_dst_array, dst_ds_name):
    return gdal.BuildVRT(dst_ds_name, src_dst_array)


def validate_band_count(src_ds, count=4):
    if src_ds.RasterCount == count:
        return src_ds
    return None
